import os
from typing import Optional, List, Tuple, Callable, Generator
from inspect import getfullargspec

# Get a filename format
def get_format (filename : str) -> str:
    if not filename:
        return None
    return filename.split('.')[-1]

# Find a function which is suitable for any of the available request "format sets"**
# All functions are checked for each request format set before jumping to another and they are evaluated in order
# A function and new generated format set with formats in common are returned
# None is returned when there is no suitable function
# WARNING: Available functions must have the 'format_sets' property
# ** Format sets are dictionaries which specify input and output formats
# Consider function format sets as 'required input' and 'available output'
# Consider request format sets as 'available input' and 'required output'
# Both inputs and ouputs are dictionaries where keys are function arguments and values are sets of supported formats
# Alternatively, an argument may have None as value to represent unnecessary requirements or missing availabilities
# An example is shown below:
# {
#     'inputs': {
#         'input_structure_filename': {'tpr'},
#         'input_trajectory_filenames': {'xtc', 'trr'},
#     },
#     'outputs': {
#         'output_trajectory_filename': {'pdb', 'gro'}
#     },
# }
# Functions may have multiple format sets since different input formats may lead to different output formats
def get_format_set_suitable_function (
    available_functions : List[Callable],
    available_request_format_sets : List[dict],
) -> Generator[ Optional[ Tuple[ Callable, dict ]], None, None ]:
    # Try with each request format set
    for request_format_set in available_request_format_sets:
        # Search functions to match formats for every required argument
        for function in available_functions:
            # Test every function format set independently
            for function_format_set in function.format_sets:
                # Check format keys are compatible
                if not check_format_sets_compability(request_format_set, function_format_set):
                    raise ValueError('Format keys are not compatible with function ' + str(function.__name__))
                # Check the function inputs to be fulfilled by the request inputs
                required_inputs = function_format_set.get('inputs', None)
                available_inputs = request_format_set.get('inputs', None)
                common_inputs = get_common_argument_formats(required_inputs, available_inputs)
                # If any of the common format sets was empty it means formats do not match
                if not common_inputs:
                    continue
                # Check the request outputs to be fulfilled by the function outputs
                required_outputs = request_format_set.get('outputs', None)
                available_outputs = function_format_set.get('outputs', None)
                common_outputs = get_common_argument_formats(required_outputs, available_outputs)
                # If any of the common format sets was empty it means formats do not match
                if not common_outputs:
                    continue
                # Generate a new format set with the common formats for every argument
                common_format_set = { 'inputs': common_inputs, 'outputs': common_outputs }
                # Otherwise we have the function
                yield function, common_format_set

# Get compatible formats between two groups of arguments
# All required argument formats must be fulfilled by the available argument formats
# Arguments are defined as dictionaries with argument names as keys and sets of available formats as values
# e.g. {
#     'input_structure_filename': {'tpr'},
#     'input_trajectory_filenames': {'xtc', 'trr'},
# },
def get_common_argument_formats (required_arguments : dict, available_arguments : dict):
    # If there are not required arguments we return an empty dictionary
    if not required_arguments:
        return {}
    # If there are not available arguments we return None
    if not available_arguments:
        return None
    # Set a dictionary with the same keys that the input arguments including only the common formats
    common_argument_formats = {}
    # Iterate over each required argument
    for required_argument, required_formats in required_arguments.items():
        # If there is not format required for this argument we set None for this argument common formats
        if required_formats == None:
            common_argument_formats[required_argument] = None
            continue
        # Get the available formats for the required input
        available_formats = available_arguments.get(required_argument, None) # DANI cuidao aquí, que igual el get no hacía falta
        # If the available arguments are missing this required argument then the function is not compatible
        if available_formats == None:
            return None
        # Find the formats in common between both the required and the available arguments
        common_formats = available_formats.intersection(required_formats)
        # If the common formats set is empty for this required argument then the function is not compatible
        if not common_formats:
            return None
        common_argument_formats[required_argument] = common_formats
    return common_argument_formats

# Check two format sets to be compatible
# Both function and request format sets must match in their requirements
# i.e. all function format set input arguments must be included in request format set input arguments
# i.e. all request format set output arguments must be included in function format set output arguments
def check_format_sets_compability (request_format_set : dict, function_format_set : dict) -> bool:
    # Check the function inputs keyowrds to exist in the request input arguments
    required_inputs = function_format_set.get('inputs', None)
    if required_inputs:
        required_input_arguments = required_inputs.keys()
        available_inputs = request_format_set.get('inputs', None)
        if not available_inputs:
            print('ERROR: Missing inputs')
            return False
        available_input_arguments = available_inputs.keys()
        for argument in required_input_arguments:
            if argument not in available_input_arguments:
                print('ERROR: Missing ' + argument + ' argument')
                return False
    # Check the request output keyowrds to exist in the function output arguments
    required_outputs = request_format_set.get('outputs', None)
    if required_outputs:
        required_output_arguments = required_outputs.keys()
        available_outputs = function_format_set.get('outputs', None)
        if not available_outputs:
            print('ERROR: Missing outputs')
            return False
        available_output_arguments = function_format_set['outputs'].keys()
        for argument in required_output_arguments:
            if argument not in available_output_arguments:
                print('ERROR: Missing ' + argument + ' argument')
                return False
    return True




# WARNING: This function makes only sets for those functions whose output can be reused as input of others
# WARNING: i.e. functions which return structure/trajectory files
# WARNING: This function should be called only when "get_format_set_suitable_function" has failed
def get_format_set_suitable_combination (
    available_functions : List[Callable],
    available_request_format_sets : List[dict],
) -> Optional[ Tuple[ Callable, dict ] ]:
    
    # Try with each request format set
    for request_format_set in available_request_format_sets:

        # Get the required outputs
        required_outputs = request_format_set.get('outputs', None)
    
        # For each function + format set possibility which is compatible with the required inputs, return the available outputs
        def get_combinations (
            current_functions : List[Callable],
            current_function_common_inputs : List[dict],
            available_inputs : dict,
        ):
            # Search functions to match formats for every required argument
            for function in available_functions:
                # Test every function format set independently
                for function_format_set in function.format_sets:
                    # Check format keys are compatible
                    if not check_format_sets_compability(request_format_set, function_format_set):
                        raise ValueError('Format keys are not compatible with function ' + str(function.__name__))
                    # We add the current function to the list of functions to combine
                    new_functions = [ *current_functions, function ]
                    # Check the function inputs to be fulfilled by the request inputs
                    required_inputs = function_format_set.get('inputs', None)
                    common_inputs = get_common_argument_formats(required_inputs, available_inputs)
                    new_function_common_inputs = [ *current_function_common_inputs, common_inputs ]
                    # If any of the common format sets was empty it means formats do not match
                    if not common_inputs:
                        continue
                    # Check the request outputs to be fulfilled by the function outputs
                    available_outputs = function_format_set.get('outputs', None)
                    common_outputs = get_common_argument_formats(required_outputs, available_outputs)
                    # If any of the common format sets was not empty then we have found a successful combination
                    if common_outputs:
                        # Use this print to display which functions are selected
                        #print([ cf.__name__ + ' from ' + cf.__module__ for cf in new_functions ])
                        yield new_functions, new_function_common_inputs, common_outputs
                    # If we have no common outputs yet we must make another jump
                    # Merge current available inputs with the current function available outputs
                    # The using all these formats as available inputs, try to find another function
                    current_structure_inputs = available_inputs.get('input_structure_filename', set())
                    current_trajectory_inputs = available_inputs.get('input_trajectory_filenames', set())
                    new_structure_inputs = available_outputs.get('output_structure_filename', set())
                    new_trajectory_inputs = available_outputs.get('output_trajectory_filename', set())
                    next_structure_inputs = current_structure_inputs.union(new_structure_inputs)
                    next_trajectory_inputs = current_trajectory_inputs.union(new_trajectory_inputs)
                    # If current function did not add new formats to the already available formats then current step was useless
                    # In this case, stop here
                    if len(next_structure_inputs) == len(current_structure_inputs) and len(next_trajectory_inputs) == len(current_trajectory_inputs):
                        continue
                    # Build the new avaiable inputs dictionary
                    new_available_inputs = {
                        'input_structure_filename': next_structure_inputs,
                        'input_trajectory_filenames': next_trajectory_inputs
                    }
                    # In case we have new available input formats we find a a new function to get the final desired output format
                    for results in get_combinations(new_functions, new_function_common_inputs, new_available_inputs):
                        if results:
                            yield results
        
        # Get every possible combination after combining the corresponding functions
        first_available_inputs = request_format_set.get('inputs', None)
        for functions, function_common_inputs, last_common_outputs in get_combinations([], [], first_available_inputs):
            # Combine all functions into one single function
            def combined_function (
                input_structure_filename : Optional[str] = None,
                input_trajectory_filenames : Optional[List[str]] = None,
                output_structure_filename : Optional[str] = None,
                output_trajectory_filename : Optional[str] = None
            ):
                auxiliar_filenames = []
                available_structure_filenames = [ input_structure_filename ]
                available_trajectory_filenames = [ input_trajectory_filenames ] # This is a list of lists
                current_input_structure_filename = input_structure_filename
                current_input_trajectory_filenames = input_trajectory_filenames
                current_output_structure_filename = None
                current_output_trajectory_filename = None
                functions_count = len(functions)
                for i, function in enumerate(functions):
                    # Get the next function common inputs in order to know what format we must output
                    next_function_index = i + 1
                    already_existing_structure = None
                    already_existing_trajectories = None
                    if next_function_index < functions_count:
                        next_function_common_inputs = function_common_inputs[next_function_index]
                        # Find the formats for the outputs. Use the first common format to do so
                        # First select the structure format
                        next_function_common_structure_formats = next_function_common_inputs.get('input_structure_filename', None)
                        if next_function_common_structure_formats:
                            output_structure_format = list(next_function_common_structure_formats)[0]
                            # Set the output structure filename
                            # Set the ouput as None if there is a structure with the desired format already
                            # Otherwise, create it using an auxiliar filename
                            already_existing_structure = next(
                                ( structure for structure in available_structure_filenames if get_format(structure) == output_structure_format ),
                                None
                            )
                            if already_existing_structure:
                                current_output_structure_filename = None
                            else:
                                auxiliar_structure_filename = '.structure.' + output_structure_format
                                current_output_structure_filename = auxiliar_structure_filename
                                auxiliar_filenames.append(auxiliar_structure_filename)
                        else:
                            current_output_structure_filename = None
                        # Then select the trajectory format
                        next_function_common_trajectory_formats = next_function_common_inputs.get('input_trajectory_filenames', None)
                        if next_function_common_trajectory_formats:
                            output_trajectory_format = list(next_function_common_trajectory_formats)[0]
                            # Set the output trajectory filenames
                            # Set the ouput as None if there are trajectories with the desired format already
                            # Otherwise, create it using an auxiliar filename
                            already_existing_trajectories = next(
                                ( trajectories for trajectories in available_trajectory_filenames if get_format(trajectories[0]) == output_trajectory_format ),
                                None
                            )
                            if already_existing_trajectories:
                                current_output_trajectory_filename = None
                            else:
                                auxiliar_trajectory_filename = '.trajectory.' + output_trajectory_format
                                current_output_trajectory_filename = auxiliar_trajectory_filename
                                auxiliar_filenames.append(auxiliar_trajectory_filename)
                        else:
                            current_output_trajectory_filename = None
                    # In case this is the last function use the final output filenames
                    else:
                        current_output_structure_filename = output_structure_filename
                        current_output_trajectory_filename = output_trajectory_filename
                    # Set the arguments to be passed to the function
                    # This has to be anticipated since we cannot pass an argument the function does not expect
                    converting_function_arguments = getfullargspec(function)[0]
                    passing_arguments = {}
                    if 'input_structure_filename' in converting_function_arguments:
                        passing_arguments['input_structure_filename'] = current_input_structure_filename
                    if 'input_trajectory_filenames' in converting_function_arguments:
                        passing_arguments['input_trajectory_filenames'] = current_input_trajectory_filenames
                    if 'output_structure_filename' in converting_function_arguments:
                        passing_arguments['output_structure_filename'] = current_output_structure_filename
                    if 'output_trajectory_filename' in converting_function_arguments:
                        passing_arguments['output_trajectory_filename'] = current_output_trajectory_filename
                    # Excute the current function
                    function(**passing_arguments)
                    # Now set the inputs for the next function
                    # Also update the available structure/trajectory files in case a further function wants to reuse them
                    if already_existing_structure:
                        current_input_structure_filename = already_existing_structure
                    else:
                        current_input_structure_filename = current_output_structure_filename
                        available_structure_filenames.append(current_output_structure_filename)
                    if already_existing_trajectories:
                        current_input_trajectory_filenames = already_existing_trajectories
                    else:
                        current_input_trajectory_filenames = [ current_output_trajectory_filename ]
                        available_trajectory_filenames.append([ current_output_trajectory_filename ])
                # Remove auxililar files
                for auxiliar_filename in auxiliar_filenames:
                    os.remove(auxiliar_filename)

            # Set the combined function format set
            combined_format_set = {
                'inputs': first_available_inputs,
                'outputs': last_common_outputs
            }
            combined_function.format_sets = [combined_format_set]

            yield combined_function, combined_format_set




# Structure file formats

def is_pdb (filename : str) -> bool:
    return filename[-4:] == '.pdb'

def is_psf (filename : str) -> bool:
    return filename[-4:] == '.psf'

def is_tpr (filename : str) -> bool:
    return filename[-4:] == '.tpr'

def is_gro (filename : str) -> bool:
    return filename[-4:] == '.gro'

def is_prmtop (filename : str) -> bool:
    return filename[-7:] == '.prmtop'

def is_top (filename : str) -> bool:
    return filename[-4:] == '.top'

# Trajectory file formats

def is_xtc (filename : str) -> bool:
    return filename[-4:] == '.xtc'

def is_dcd (filename : str) -> bool:
    return filename[-4:] == '.dcd'

def is_netcdf (filename : str) -> bool:
    return filename[-3:] == '.nc'

def are_xtc (filenames : list) -> bool:
    return all([ is_xtc(filename) for filename in filenames ])

def are_dcd (filenames : list) -> bool:
    return all([ is_dcd(filename) for filename in filenames ])

def are_netcdf (filenames : list) -> bool:
    return all([ is_netcdf(filename) for filename in filenames ])

# Extra formats logic

# Check if a file may be read by pytraj according to its format
def is_pytraj_supported (filename : str) -> bool:
    return is_prmtop(filename) or is_top(filename) or is_psf(filename)

# From GitHub:
# ParmFormatDict = {
#     "AMBERPARM": AMBERPARM,
#     "PDBFILE": PDBFILEPARM,
#     "MOL2FILE": MOL2FILEPARM,
#     "CHARMMPSF": CHARMMPSF,
#     "CIFFILE": CIFFILE,
#     "GMXTOP": GMXTOP,
#     "SDFFILE": SDFFILE,
#     "TINKER": TINKERPARM,
#     "UNKNOWN_PARM": UNKNOWN_PARM,
# }

# Get the pytraj format key for the write_parm function for a specific file according to its format
def get_pytraj_parm_format (filename : str) -> str:
    if is_prmtop(filename):
        return 'AMBERPARM'
    if is_psf(filename):
        return 'CHARMMPSF'
    if is_top(filename):
        return 'GMXTOP'
    if is_pdb(filename):
        return 'PDBFILE'
    raise ValueError('The file ' + filename + ' format is not supported')


from typing import List, Set, Optional

from .formats import get_format, get_format_set_suitable_function
from .gmx_spells import get_first_frame as gmx_get_first_frame
from .mdt_spells import get_first_frame as mdt_get_first_frame
from .mdt_spells import get_last_frame as mdt_get_last_frame

# Set the functions to perform single frame gettering
first_frame_getter_functions = [ gmx_get_first_frame, mdt_get_first_frame ]
last_frame_getter_functions = [ mdt_get_last_frame ]

# Get the first frame from a trajectory
# Return the single frame filename
# WARNING: The structure filename may be None
def get_first_frame (
    input_structure_filename : str,
    input_trajectory_filename : str,
    accepted_output_formats : Optional[Set[str]] = None
) -> str:
    # Get the input formats
    input_structure_format = get_format(input_structure_filename)
    input_trajectory_format = get_format(input_trajectory_filename)
    format_set = {
        'inputs': {
            'input_structure_filename': None if input_structure_format == None else { input_structure_format },
            'input_trajectory_filename': { input_trajectory_format }
        },
        'outputs': {
            'output_frame_filename': accepted_output_formats
        }
    }
    # Get a suitable function to obtain the unique frame
    suitables = next(get_format_set_suitable_function(
        available_functions=first_frame_getter_functions,
        available_request_format_sets=[format_set],
    ), None)
    if not suitables:
        raise SystemExit('There is no first frame getter function which supports the requested formats')
    suitable_function, formats = suitables
    # The output format will be the first common format between the available formats and the function formats
    output_format = list(formats['outputs']['output_frame_filename'])[0]
    # Set the output filename
    output_single_frame_filename = '.single_frame.' + output_format
    suitable_function(input_structure_filename, input_trajectory_filename, output_single_frame_filename)
    return output_single_frame_filename

# Get the last frame from a trajectory
# Return the single frame filename
# WARNING: The structure filename may be None
def get_last_frame (
    input_structure_filename : str,
    input_trajectory_filename : str,
    accepted_output_formats : Optional[Set[str]] = None
) -> str:
    # Get the input formats
    input_structure_format = get_format(input_structure_filename)
    input_trajectory_format = get_format(input_trajectory_filename)
    format_set = {
        'inputs': {
            'input_structure_filename': None if input_structure_format == None else { input_structure_format },
            'input_trajectory_filename': { input_trajectory_format }
        },
        'outputs': {
            'output_frame_filename': accepted_output_formats
        }
    }
    # Get a suitable function to obtain the unique frame
    suitables = next(get_format_set_suitable_function(
        available_functions=last_frame_getter_functions,
        available_request_format_sets=[format_set],
    ), None)
    if not suitables:
        raise SystemExit('There is no last frame getter function which supports the requested formats')
    suitable_function, formats = suitables
    # The output format will be the last common format between the available formats and the function formats
    output_format = list(formats['outputs']['output_frame_filename'])[0]
    # Set the output filename
    output_single_frame_filename = '.single_frame.' + output_format
    suitable_function(input_structure_filename, input_trajectory_filename, output_single_frame_filename)
    return output_single_frame_filename



import pytraj as pyt

# Set pytraj supported formats
pytraj_supported_structure_formats = {'prmtop', 'pdb', 'parm7', 'mol2', 'psf', 'cif', 'top', 'sdf'}
pytraj_supported_trajectory_formats = {'xtc', 'trr', 'crd', 'nc', 'dcd'}

# Get the trajectory frames number using pytraj
def get_frames_count (
    input_topology_filename : str,
    input_trajectory_filename : str) -> int:
    
    # Load the trajectory from pytraj
    pyt_trajectory = pyt.iterload(
        input_trajectory_filename,
        input_topology_filename)

    # Return the frames number
    return pyt_trajectory.n_frames
# Set function supported formats
get_frames_count.format_sets = [
    {
        'inputs': {
            'input_structure_filename': pytraj_supported_structure_formats,
            'input_trajectory_filename': pytraj_supported_trajectory_formats
        }
    }
]


from argparse import ArgumentParser, RawTextHelpFormatter

from .conversions import convert
from .vmd_spells import chainer
from .mdt_spells import split_merged_trajectories, get_trajectory_subset

# Define console call for mdtoolbelt
parser = ArgumentParser(description="Call a tool from mdtoolbelt", formatter_class=RawTextHelpFormatter)
subparsers = parser.add_subparsers(help='Name of the tool to be used', dest="tool")

# The convert command
convert_parser = subparsers.add_parser("convert")
convert_parser.add_argument(
    "-is", "--input_structure",
    help="Path to input structure file")
convert_parser.add_argument(
    "-os", "--output_structure",
    help="Path to output structure file")
convert_parser.add_argument(
    "-it", "--input_trajectories", nargs='*',
    help="Path to input trajectory file(s)")
convert_parser.add_argument(
    "-ot", "--output_trajectory",
    help="Path to output trajectory file")

# The chainer command
chainer_parser = subparsers.add_parser("chainer")
chainer_parser.add_argument(
    "-is", "--input_structure", required=True,
    help="Path to input structure file (pdb)")
chainer_parser.add_argument(
    "-sel", "--atom_selection",
    help="Atoms to be chained")
chainer_parser.add_argument(
    "-chn", "--chain_letter",
    help="The chain letter to be set in selected atoms")
chainer_parser.add_argument(
    "-os", "--output_structure",
    help="Path to output structure file (pdb)")

# The split command
split_parser = subparsers.add_parser("split")
split_parser.add_argument(
    "-is", "--input_structure", required=True,
    help="Path to input structure file")
split_parser.add_argument(
    "-it", "--input_trajectory",
    help="Path to input trajectory file")
split_parser.add_argument(
    "-cut", "--cutoff", type=float, default=0.2,
    help="The minimum size of the RMSD jump to consider it is a different trajectory")
split_parser.add_argument(
    "-otp", "--output_trajectory_prefix", default="split",
    help="Prefix for the path to output trajectory files")

# The subset command
subset_parser = subparsers.add_parser("subset")
subset_parser.add_argument(
    "-is", "--input_structure", required=True,
    help="Path to input structure file")
subset_parser.add_argument(
    "-it", "--input_trajectory",
    help="Path to input trajectory file")
subset_parser.add_argument(
    "-ot", "--output_trajectory",
    help="Path to output trajectory file")
subset_parser.add_argument(
    "-start", "--start", type=int, default=0,
    help="Start frame")
subset_parser.add_argument(
    "-end", "--end", type=int, default=0,
    help="End frame")
subset_parser.add_argument(
    "-step", "--step", type=int, default=1,
    help="Frame step")

args = parser.parse_args()

def call():

    # If no command is passed print help
    tool = args.tool
    if not tool:
        parser.print_help()
        return

    # In case the convert tool was called
    if tool == 'convert':
        # If no input arguments are passed print help
        if args.input_structure == None and args.input_trajectories == None:
            convert_parser.print_help()
            return
        # Run the convert command
        convert(
            input_structure_filename=args.input_structure,
            output_structure_filename=args.output_structure,
            input_trajectory_filenames=args.input_trajectories,
            output_trajectory_filename=args.output_trajectory,
        )

    # In case the chainer tool was called
    if tool == 'chainer':
        chainer(
            input_pdb_filename=args.input_structure,
            atom_selection=args.atom_selection,
            chain_letter=args.chain_letter,
            output_pdb_filename=args.output_structure
        )

    if tool == 'split':
        split_merged_trajectories(
            input_structure_filename=args.input_structure,
            input_trajectory_filename=args.input_trajectory,
            sudden_jump_cutoff=args.cutoff,
            output_trajectory_prefix=args.output_trajectory_prefix
        )

    if tool == 'subset':
        get_trajectory_subset(
            input_structure_filename=args.input_structure,
            input_trajectory_filename=args.input_trajectory,
            output_trajectory_filename=args.output_trajectory,
            start=args.start,
            end=args.end,
            step=args.step
        )

    # Tool will always match one of the previous defined options
    # Otherwise argparse returns error itself
    print('Done :)')


from .formats import get_format, get_format_set_suitable_function
from .pyt_spells import get_frames_count as pyt_get_frames_count

# Set the functions to perform trajectory frame count
frame_count_functions = [ pyt_get_frames_count ]

# Get the trajectory frame count
# WARNING: The structure filename may be None
def get_frames_count (
    input_structure_filename : str,
    input_trajectory_filename : str,
) -> int:
    # Get the input formats
    input_structure_format = get_format(input_structure_filename)
    input_trajectory_format = get_format(input_trajectory_filename)
    format_set = {
        'inputs': {
            'input_structure_filename': None if input_structure_format == None else { input_structure_format },
            'input_trajectory_filename': { input_trajectory_format }
        }
    }
    # Get a suitable function
    suitables = next(get_format_set_suitable_function(
        available_functions=frame_count_functions,
        available_request_format_sets=[format_set],
    ), None)
    if not suitables:
        raise SystemExit('There is no frame count function which supports the requested formats')
    suitable_function, formats = suitables
    # Run the suitable function
    frame_count = suitable_function(input_structure_filename, input_trajectory_filename)
    return frame_count


# Functions powered by VMD
# Humphrey, W., Dalke, A. and Schulten, K., "VMD - Visual Molecular Dynamics", J. Molec. Graphics, 1996, vol. 14, pp. 33-38. 
# http://www.ks.uiuc.edu/Research/vmd/

import os

from subprocess import run, PIPE, STDOUT, Popen
from typing import Optional, List

from .single_frame_getter import get_last_frame
from .formats import get_format

# Set the script filename with all commands to be passed to vmd
commands_filename = '.commands.vmd'

# List all the vmd supported trajectory formats
vmd_supported_structure_formats = {'pdb', 'prmtop', 'psf', 'gro'} # DANI: Esto lo he hecho rápido, hay muchas más
vmd_supported_trajectory_formats = {'mdcrd', 'crd', 'dcd', 'xtc', 'trr', 'nc'}

# Given a vmd supported topology with no coordinates and a single frame file, generate a pdb file
def vmd_to_pdb (
    input_structure_filename : str,
    input_trajectory_filename : str,
    output_structure_filename : str):

    # DANI: Esto alomejor se podria substituit por un 'animate read' en los commands
    # https://www.ks.uiuc.edu/Research/vmd/current/ug/node121.html
    single_frame_filename = get_last_frame(input_structure_filename, input_trajectory_filename, vmd_supported_trajectory_formats)

    # Prepare a script for VMD to run. This is Tcl language
    with open(commands_filename, "w") as file:
        # Select all atoms
        file.write('set all [atomselect top "all"]\n')
        # Write the current topology in 'pdb' format
        file.write('$all frame first\n')
        file.write('$all writepdb ' + output_structure_filename + '\n')
        file.write('exit\n')

    # Run VMD
    logs = run([
        "vmd",
        input_structure_filename,
        single_frame_filename,
        "-e",
        commands_filename,
        "-dispdev",
        "none"
    ], stdout=PIPE).stdout.decode()

    os.remove(commands_filename)
# Set function supported formats
vmd_to_pdb.format_sets = [
    {
        'inputs': {
            'input_structure_filename': {'psf', 'parm', 'prmtop'},
            'input_trajectory_filename': vmd_supported_trajectory_formats
        },
        'outputs': {
            'output_structure_filename': {'pdb'}
        }
    }
]

# This tool allows you to set the chain of all atoms in a selection
# This is powered by VMD and thus the selection lenguage must be the VMD's
# Arguments are as follows:
# 1 - Input pdb filename
# 2 - Atom selection (All atoms by defualt)
# 3 - Chain letter (May be the flag 'fragment', which is the default indeed)
# 4 - Output pdb filename (Input filename by default)
# WARNING: When no selection is passed, if only a part of a "fragment" is missing the chain then the whole fragment will be affected
# WARNING: VMD only handles fragments if there are less fragments than letters in the alphabet
def chainer (
    input_pdb_filename : str,
    atom_selection : Optional[str] = None,
    chain_letter : Optional[str] = None,
    output_pdb_filename : Optional[str] = None
):

    # If no atom selection is provided then all atoms are selected
    if not atom_selection:
        atom_selection = 'all'

    # If no chain letter is provided then the flag 'fragment' is used
    if not chain_letter:
        chain_letter = 'fragment'

    # If no output filename is provided then use input filename as output filename
    if not output_pdb_filename:
        output_pdb_filename = input_pdb_filename

    # Check the file exists
    if not os.path.exists(input_pdb_filename):
        raise SystemExit('ERROR: The file does not exist')
       
    with open(commands_filename, "w") as file:
        # Select the specified atoms and set the specified chain
        file.write('set atoms [atomselect top "' + atom_selection + '"]\n')
        # In case chain letter is not a letter but the 'fragment' flag, asign chains by fragment
        # Fragments are atom groups which are not connected by any bond
        if chain_letter == 'fragment':
            # Get all different chain names
            file.write('set chains_sample [lsort -unique [${atoms} get chain]]\n')
            # Set letters in alphabetic order
            file.write('set letters "A B C D E F G H I J K L M N O P Q R S T U V W X Y Z"\n')
            # Get the number of fragments
            file.write('set fragment_number [llength [lsort -unique -integer [${atoms} get fragment]]]\n')
            # For each fragment, set the chain of all atoms which belong to this fragment alphabetically
            # e.g. fragment 0 -> chain A, fragment 1 -> chain B, ...
            file.write('for {set i 0} {$i <= $fragment_number} {incr i} {\n')
            file.write('	set fragment_atoms [atomselect top "fragment $i"]\n')
            file.write('	$fragment_atoms set chain [lindex $letters $i]\n')
            file.write('}\n')
            # Otherwise, set the specified chain
        else:
            file.write('$atoms set chain ' + chain_letter + '\n')
        # Write the current topology in 'pdb' format
        file.write('set all [atomselect top "all"]\n')
        file.write('$all frame first\n')
        file.write('$all writepdb ' + output_pdb_filename + '\n')
        file.write('exit\n')

    # Run VMD
    logs = run([
        "vmd",
        input_pdb_filename,
        "-e",
        commands_filename,
        "-dispdev",
        "none"
    ], stdout=PIPE).stdout.decode()

    # Remove the vmd commands file
    os.remove(commands_filename)

# Get vmd supported trajectories merged and converted to a different format
# WARNING: Note that this process is not memory efficient so beware the size of trajectories to be converted
# WARNING: The input structure filename may be None
def merge_and_convert_trajectories (
    input_structure_filename : Optional[str],
    input_trajectory_filenames : List[str],
    output_trajectory_filename : str
    ):

    # Get the format to export coordinates
    output_trajectory_format = get_format(output_trajectory_filename)

    # Although 'crd' and 'mdcrd' may be the same, VMD only recognizes 'crd' as exporting coordinates file type
    if output_trajectory_format == 'mdcrd':
        output_trajectory_format = 'crd'

    # Prepare a script for the VMD to automate the data parsing. This is Tcl lenguage
    # In addition, if chains are missing, this script asigns chains by fragment
    # Fragments are atom groups which are not connected by any bond
    with open(commands_filename, "w") as file:
        # Select all atoms
        file.write('set all [atomselect top "all"]\n')
        # Write the current trajectory in the specified format format
        file.write('animate write ' + output_trajectory_format + ' ' + output_trajectory_filename + ' waitfor all sel $all\n')
        file.write('exit\n')

    inputs = [ input_structure_filename, *input_trajectory_filenames ] if input_structure_filename else input_trajectory_filenames

    # Run VMD
    logs = run([
        "vmd",
        *inputs,
        "-e",
        commands_filename,
        "-dispdev",
        "none"
    ], stdout=PIPE, stderr=STDOUT).stdout.decode() # Redirect errors to the output in order to dont show them in console

    if not os.path.exists(output_trajectory_filename):
        print(logs)
        raise SystemExit('Something went wrong with VMD')

    # Remove the vmd commands file
    os.remove(commands_filename)

# Set function supported formats
merge_and_convert_trajectories.format_sets = [
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filenames': {'dcd', 'xtc', 'trr', 'nc'}
        },
        'outputs': {
            # NEVER FORGET: VMD cannot write to all formats it supports to read
            'output_trajectory_filename': {'mdcrd', 'crd', 'dcd', 'trr'}
        }
    },
    {
        'inputs': {
            'input_structure_filename': vmd_supported_structure_formats,
            'input_trajectory_filenames': {'mdcrd', 'crd'}
        },
        'outputs': {
            # NEVER FORGET: VMD cannot write to all formats it supports to read
            'output_trajectory_filename': {'mdcrd', 'crd', 'dcd', 'trr'}
        }
    }
]


import os
from shutil import copyfile
from subprocess import run, PIPE, Popen

from typing import List

from .formats import get_format

# Get the first frame from a trajectory
def get_first_frame (input_structure_filename : str, input_trajectory_filename : str, output_frame_filename : str):
    # Run Gromacs
    if input_structure_filename:
        p = Popen([
            "echo",
            "System",
        ], stdout=PIPE)
        logs = run([
            "gmx",
            "trjconv",
            "-s",
            input_structure_filename,
            "-f",
            input_trajectory_filename,
            "-o",
            output_frame_filename,
            "-dump",
            "0",
            "-quiet"
        ], stdin=p.stdout, stderr=PIPE).stderr.decode()
    else:
        logs = run([
            "gmx",
            "trjconv",
            "-f",
            input_trajectory_filename,
            "-o",
            output_first_frame_filename,
            "-dump",
            "0",
            "-quiet"
        ], stderr=PIPE).stderr.decode()
    # If output has not been generated then warn the user
    if not os.path.exists(output_frame_filename):
        print(logs)
        raise SystemExit('Something went wrong with Gromacs')

# Set function supported formats
get_first_frame.format_sets = [
    {
        'inputs': {
            'input_structure_filename': {'tpr', 'pdb', 'gro'},
            'input_trajectory_filename': {'xtc', 'trr'}
        },
        'outputs': {
            'output_frame_filename': {'pdb', 'gro'}
        }
    },
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filename': {'xtc', 'trr'}
        },
        'outputs': {
            'output_frame_filename': {'xtc', 'trr'}
        }
    },
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filename': {'pdb'}
        },
        'outputs': {
            'output_frame_filename': {'pdb', 'xtc', 'trr'}
        }
    },
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filename': {'gro'}
        },
        'outputs': {
            'output_frame_filename': {'gro', 'xtc', 'trr'}
        }
    }
]

# Get the structure of a tpr file using the first frame getter function
def get_tpr_structure (input_structure_filename : str, input_trajectory_filename : str, output_structure_filename : str):
    get_first_frame(input_structure_filename, input_trajectory_filename, output_structure_filename)
get_tpr_structure.format_sets = [
    {
        'inputs': {
            'input_structure_filename': {'tpr', 'pdb', 'gro'},
            'input_trajectory_filename': {'xtc', 'trr'}
        },
        'outputs': {
            'output_structure_filename': {'pdb', 'gro'}
        }
    }
]

# Get gromacs supported trajectories merged and converted to a different format
def merge_and_convert_trajectories (input_trajectory_filenames : List[str], output_trajectory_filename : str):
    # Get trajectory formats
    sample_trajectory = input_trajectory_filenames[0]
    input_trajectories_format = get_format(sample_trajectory)
    output_trajectory_format = get_format(output_trajectory_filename)
    auxiliar_single_trajectory_filename = '.single_trajectory.' + input_trajectories_format
    # If we have multiple trajectories then join them
    if len(input_trajectory_filenames) > 1:
        single_trajectory_filename = auxiliar_single_trajectory_filename
        logs = run([
            "gmx",
            "trjcat",
            "-f",
            *input_trajectory_filenames,
            "-o",
            single_trajectory_filename,
            "-quiet"
        ], stderr=PIPE).stderr.decode()
        # If output has not been generated then warn the user
        if not os.path.exists(single_trajectory_filename):
            print(logs)
            raise SystemExit('Something went wrong with Gromacs')
    else:
        single_trajectory_filename = sample_trajectory
    # In case input and output formats are different we must convert the trajectory
    if input_trajectories_format != output_trajectory_format:
        logs = run([
            "gmx",
            "trjconv",
            "-f",
            single_trajectory_filename,
            "-o",
            output_trajectory_filename,
            "-quiet"
        ], stderr=PIPE).stderr.decode()
        # If output has not been generated then warn the user
        if not os.path.exists(output_trajectory_filename):
            print(logs)
            raise SystemExit('Something went wrong with Gromacs')
    else:
        copyfile(single_trajectory_filename, output_trajectory_filename)
    # Remove residual files
    if os.path.exists(auxiliar_single_trajectory_filename):
        os.remove(auxiliar_single_trajectory_filename)

merge_and_convert_trajectories.format_sets = [
    {
        'inputs': {
            'input_trajectory_filenames': {'xtc', 'trr'}
        },
        'outputs': {
            'output_trajectory_filename': {'xtc', 'trr'}
        }
    },
    {
        'inputs': {
            'input_trajectory_filenames': {'pdb'}
        },
        'outputs': {
            'output_trajectory_filename': {'pdb', 'xtc', 'trr'}
        }
    },
    {
        'inputs': {
            'input_trajectory_filenames': {'gro'}
        },
        'outputs': {
            'output_trajectory_filename': {'gro', 'xtc', 'trr'}
        }
    }
]


from typing import List

# A selection is a list of atom indices from a structure
class Selection:

    def __init__ (self, atom_indices : List[int]):
        self.atom_indices = atom_indices

    def __repr__ (self):
        return '<Selection (' + str(len(self.atom_indices)) + ' atoms)>'

    def __len__ (self):
        return len(self.atom_indices)

    @classmethod
    def from_prody (cls, prody_selection):
        indexes = [ atom.getIndex() for atom in prody_selection.iterAtoms() ]
        return cls(indexes)

    def to_prody (self) -> str:
        return 'index ' + ' '.join([ str(index) for index in self.atom_indices ])

    def to_mdanalysis (self) -> str:
        return self.to_prody() # Same notation than prody for atom indices

    def to_pytraj (self) -> str:
        # NEVER FORGET: Pytraj counts atoms starting at 1, not at 0
        return '@' + ','.join([ str(index+1) for index in self.atom_indices ])

    # Get a string made of all indexes separated by underscores
    # This string can be then passed as a bash argument and easily parsed by other programms
    # Indices can start from 0 or from 1
    def to_bash (self, one_start : bool = False) -> str:
        if one_start:
            return '_'.join([ str(index + 1) for index in self.atom_indices ])
        else:
            return '_'.join([ str(index) for index in self.atom_indices ])

    # Generate a vmd selection in tcl format
    def to_vmd (self) -> str:
        return 'index ' + ' '.join([ str(index) for index in self.atom_indices ])

    # Generate the file content of gromacs ndx file
    def to_ndx (self, selection_name : str = 'Selection') -> str:
        # Add a header 
        content = '[ ' + selection_name + ' ]\n'
        count = 0
        for index in self.atom_indices:
            # Add a breakline each 15 indices
            count += 1
            if count == 15:
                content += '\n'
                count = 0
            # Add a space between indices
            # Atom indices go from 0 to n-1
            # Add +1 to the index since gromacs counts from 1 to n
            content += str(index + 1) + ' '
        content += '\n' 
        return content


import os
from shutil import copyfile
from typing import List, Optional
from inspect import getfullargspec

from .formats import get_format, get_format_set_suitable_function, get_format_set_suitable_combination
from .vmd_spells import vmd_to_pdb
from .gmx_spells import get_tpr_structure
from .gmx_spells import merge_and_convert_trajectories as gmx_merge_and_convert_trajectories
from .mdt_spells import merge_and_convert_trajectories as mdt_merge_and_convert_trajectories
from .vmd_spells import merge_and_convert_trajectories as vmd_merge_and_convert_trajectories

# Set functions to performe structure conversions
# These functions must have 'input_structure_filename' and 'output_structure_filename' keywords
# These functions must have the 'format_sets' property
# These functions may have the 'input_trajectory_filename' keyword
structure_converting_functions = [ get_tpr_structure, vmd_to_pdb ]

# Set functions to performe trajectory conversions
# These functions must have 'input_trajectory_filename' and 'output_trajectory_filename' keywords
# These functions must have the 'format_sets' property
trajectory_converting_functions = [
    mdt_merge_and_convert_trajectories,
    gmx_merge_and_convert_trajectories,
    vmd_merge_and_convert_trajectories
]

# Handle conversions of different structure and trajectory formats
# Merge multiple input trajectories into one single output trajectory
# Inputs are the original strucutre and/or trajectory files and the list of possible output filenames
# Only one of each group of output filenames will be generated (if possible)
# Return the names of the generated output files
def convert (
    input_structure_filename :  Optional[str] = '',
    output_structure_filename : Optional[str] = '',
    input_trajectory_filenames :  Optional[List[str]] = '',
    output_trajectory_filename : Optional[str] = ''
):

    # If we have output but not input we must complain
    if output_structure_filename and not input_structure_filename:
        raise SystemExit('Missing input structure')
    if output_trajectory_filename and not input_trajectory_filenames:
        raise SystemExit('Missing input trajectory')

    # Get the first trajectory as a sample for those processes which do not require the whole trajectory
    trajectory_sample = input_trajectory_filenames[0]

    # Get file formats
    input_structure_format = get_format(input_structure_filename) if input_structure_filename else None
    output_structure_format = get_format(output_structure_filename) if output_structure_filename else None
    input_trajectory_format = get_format(trajectory_sample) if input_trajectory_filenames else None
    output_trajectory_format = get_format(output_trajectory_filename) if output_trajectory_filename else None

    # In case multiple trajectories are passed check all of them to match in the format
    trajectory_files_count = len(input_trajectory_filenames)
    if trajectory_files_count > 1:
        for filename in input_trajectory_filenames:
            if get_format(filename) != input_trajectory_format:
                raise SystemExit('All input trajectories must have the same format')

    # Convert the structure
    # Do it inside a function just to return as soon as we are done
    def convert_structure():
        # If there is no output filename it means we have nothing to do here
        if not output_structure_filename:
            return
        # If input and output formats are the same then just copy the file with the new name
        if input_structure_format == output_structure_format:
            copyfile(input_structure_filename, output_structure_filename)
            return
        # Otherwise, we must convert
        # Choose the right conversion function according to input and output formats
        request_format_set = {
            'inputs': {
                'input_structure_filename': { input_structure_format },
                'input_trajectory_filename': { input_trajectory_format }
            },
            'outputs': {
                'output_structure_filename': { output_structure_format }
            }
        }
        suitable = next(get_format_set_suitable_function(
            available_functions=structure_converting_functions,
            available_request_format_sets=[request_format_set],
        ), None)
        # If there is no function to handle this specific conversion we stop here
        if not suitable:
            raise SystemExit('Conversion from ' + input_structure_format +
                ' to ' + output_structure_format + ' is not supported')
        converting_function, formats = suitable
        # Find the function keywords
        # This is important since some functions may need a trajectory input in addition
        converting_function_keywords = getfullargspec(converting_function)[0]
        required_trajectory = 'input_trajectory_filename' in converting_function_keywords
        if required_trajectory:
            if not input_trajectory_filenames:
                raise SystemExit('The structure input format ' + input_structure_format +
                ' is missing coordinates and the output format ' + output_structure_format +
                ' needs them. An input trajectory file is required.')
            converting_function(
                input_structure_filename=input_structure_filename,
                input_trajectory_filename=trajectory_sample,
                output_structure_filename=output_structure_filename
            )
        else:
            converting_function(
                input_structure_filename=input_structure_filename,
                output_structure_filename=output_structure_filename
            )
    convert_structure()

    def convert_trajectory ():
        # If there is no output filename it means we have nothing to do here
        if not output_trajectory_filename:
            return
        # If there is only 1 input trajectory and it has the same format that the output then just copy the file with the new name
        if trajectory_files_count == 1 and input_trajectory_format == output_trajectory_format:
            copyfile(trajectory_sample, output_trajectory_filename)
            return
        # Otherwise, we must convert
        # Choose the right conversion function according to input and output formats
        request_format_set = {
            'inputs': {
                'input_structure_filename': { input_structure_format },
                'input_trajectory_filenames': { input_trajectory_format }
            },
            'outputs': {
                'output_trajectory_filename': { output_trajectory_format }
            }
        }
        suitable = next(get_format_set_suitable_function(
            available_functions=trajectory_converting_functions,
            available_request_format_sets=[request_format_set],
        ), None)
        # If there is no function to handle this specific conversion we try to combine several functions in order to do it
        if not suitable:
            print('WARNING: There is no function to do the conversion directly. Trying to combine multiple functions...')
            suitable = next(get_format_set_suitable_combination(
                available_functions=trajectory_converting_functions,
                available_request_format_sets=[request_format_set],
            ), None)
        # If there is no function to handle this specific conversion we stop here
        if not suitable:
            raise SystemExit('Conversion from ' + input_trajectory_format +
                ' to ' + output_trajectory_format + ' is not supported')
        converting_function, formats = suitable
        # Get the input structure expected format
        expected_input_structure_formats = formats['inputs'].get('input_structure_filename', False)
        # If the function expects any fromat then pass the structure
        if expected_input_structure_formats:
            converting_function(
                input_structure_filename=input_structure_filename,
                input_trajectory_filenames=input_trajectory_filenames,
                output_trajectory_filename=output_trajectory_filename
            )
        # If the function expects None then pass None
        elif expected_input_structure_formats == None:
            converting_function(
                input_structure_filename=None,
                input_trajectory_filenames=input_trajectory_filenames,
                output_trajectory_filename=output_trajectory_filename
            )
        # If the function has not the input structure argument then do not pass it
        else:
            converting_function(
                input_trajectory_filenames=input_trajectory_filenames,
                output_trajectory_filename=output_trajectory_filename
            )
    convert_trajectory()


import os
from subprocess import run, PIPE
from typing import List

import mdtraj as mdt

from .formats import get_format
from .frame_counts import get_frames_count

# Set mdtraj supported formats
mdtraj_supported_structure_formats = {
    'pdb', 'pdb.gz' 'h5', 'lh5', 'prmtop', 'parm7', 'prm7', 'psf', 'mol2', 'hoomdxml', 'gro', 'arc', 'hdf5', 'gsd'
}
mdtraj_supported_trajectory_formats = {'dcd', 'xtc', 'trr', 'nc', 'h5', 'binpos', 'mdcrd'}

# Use mdtraj 'mdconvert' command-line script (there is no python version for this tool apparently)
# Multiple files may be selected with bash syntax
def merge_and_convert_trajectories (
    input_trajectory_filenames : List[str],
    output_trajectory_filename : str
    ):

    # Run MDtraj
    logs = run([
        "mdconvert",
        "-o",
        output_trajectory_filename,
        *input_trajectory_filenames,
    ], stderr=PIPE).stderr.decode()
    # If output has not been generated then warn the user
    if not os.path.exists(output_trajectory_filename):
        print(logs)
        raise SystemExit('Something went wrong with MDTraj')
        
# NEVER FORGET: mdconvert does not handle mdcrd format
mdconvert_supported_formats = {'dcd', 'xtc', 'trr', 'nc', 'h5', 'binpos'}
merge_and_convert_trajectories.format_sets = [
    {
        'inputs': {
            'input_trajectory_filenames': mdconvert_supported_formats
        },
        'outputs': {
            'output_trajectory_filename': mdconvert_supported_formats
        }
    },
]

# Merge and convert trajectories without the mdconvert command
# WARNING: Note that this process is not memory efficient so beware the size of trajectories to be converted
# DEPRECATED: This function was meant to convert trajectories to mdcrd, which is not supported by mdconvert
# DEPRECATED: However the export to mdcrd in mdtraj does not allow to remove the periodic box, which may be a problem
# DEPRECTAED: Use VMD instead
def merge_and_convert_trajectories_unefficient (
    input_structure_filename : str,
    input_trajectory_filenames : List[str],
    output_trajectory_filename : str,
):
    # Load all trajectories together
    sample_trajectory = input_trajectory_filenames[0]
    if input_structure_filename:
        trajectory = mdt.load(sample_trajectory, top=input_structure_filename)
    else:
        trajectory = mdt.load(sample_trajectory)
    for extra_trajectory in input_trajectory_filenames[1:]:
        if input_structure_filename:
            extra = mdt.load(sample_trajectory, top=input_structure_filename)
        else:
            extra = mdt.load(sample_trajectory)
        trajectory = mdt.join([trajectory, extra], check_topology=False)
        
    # Write the whole trajectory
    trajectory.save(output_trajectory_filename)
    
merge_and_convert_trajectories_unefficient.format_sets = [
    {
        'inputs': {
            'input_structure_filename': mdtraj_supported_structure_formats,
            'input_trajectory_filenames': mdtraj_supported_trajectory_formats
        },
        'outputs': {
            'output_trajectory_filename': mdtraj_supported_trajectory_formats
        }
    },
]

# Get specific frames from a trajectory
def get_trajectory_subset (
    input_structure_filename : str,
    input_trajectory_filename : str,
    output_trajectory_filename : str,
    start : int = 0,
    end : int = None,
    step : int = 1
):
    # End must be grater than start
    if end != None and end < start:
        raise SystemExit('End frame must be posterior to start frame')

    # Load the trajectory frame by frame and get only the desired frames
    if input_structure_filename:
        trajectory = mdt.iterload(input_trajectory_filename, top=input_structure_filename, chunk=1)
    else:
        trajectory = mdt.iterload(input_trajectory_filename, chunk=1)
    # Get the first chunk
    reduced_trajectory = None
    for i, chunk in enumerate(trajectory):
        if i == start:
            reduced_trajectory = chunk
            break
    # Get further chunks
    relative_end = end - start if end != None else None
    for i, chunk in enumerate(trajectory, 1): # Start the count at 1
        if i == relative_end:
            break
        if i % step == 0:
            reduced_trajectory = mdt.join([reduced_trajectory, chunk], check_topology=False)

    # WARNING: This print here is not just a log. DO NOT REMOVE IT
    # WARNING: It fixes an error (ValueError: unitcell angle < 0) which happens sometimes
    print(reduced_trajectory)

    # Write reduced trajectory to output file
    reduced_trajectory.save(output_trajectory_filename)
get_trajectory_subset.format_sets = [
    {
        'inputs': {
            'input_structure_filename': mdtraj_supported_structure_formats,
            'input_trajectory_filename': mdtraj_supported_trajectory_formats
        },
        'outputs': {
            'output_trajectory_filename': mdtraj_supported_trajectory_formats
        }
    },
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filename': mdtraj_supported_structure_formats
        },
        'outputs': {
            'output_trajectory_filename': mdtraj_supported_structure_formats
        }
    }
]

# Get first frame from a trajectory
def get_first_frame (
    input_structure_filename : str,
    input_trajectory_filename : str,
    output_frame_filename : str,
):
    get_trajectory_subset(input_structure_filename, input_trajectory_filename, output_frame_filename, 0)
# Get last last from a trajectory
def get_last_frame (
    input_structure_filename : str,
    input_trajectory_filename : str,
    output_frame_filename : str,
):
    frame_count = get_frames_count(input_structure_filename, input_trajectory_filename)
    get_trajectory_subset(input_structure_filename, input_trajectory_filename, output_frame_filename, frame_count-1)
# Set function supported formats
single_frame_getter_format_sets = [
    {
        'inputs': {
            'input_structure_filename': None,
            'input_trajectory_filename': mdtraj_supported_trajectory_formats
        },
        'outputs': {
            'output_frame_filename': mdtraj_supported_trajectory_formats
        }
    },
    {
        'inputs': {
            'input_structure_filename': mdtraj_supported_structure_formats,
            'input_trajectory_filename': mdtraj_supported_trajectory_formats
        },
        'outputs': {
            'output_frame_filename': { *mdtraj_supported_structure_formats, *mdtraj_supported_trajectory_formats }
        }
    },
]
get_first_frame.format_sets = single_frame_getter_format_sets
get_last_frame.format_sets = single_frame_getter_format_sets

# Split a trajectory which is actually a merge of independent trajectories back to the original pieces
# Run an RMSD analysis to guess where the pieces are
# The cutoff sets the minimum RMSD change to consider it is a different trajectory
def split_merged_trajectories (
    input_structure_filename : str,
    input_trajectory_filename : str,
    sudden_jump_cutoff : float = 0.2,
    output_trajectory_prefix : str = 'split'):
    # Get the input trajectory format
    input_trajectory_format = get_format(input_trajectory_filename)
    # The cutoff must be a negative number since independent trajectories RMSD sudden jumps will be negative
    cutoff = -abs(sudden_jump_cutoff)
    # Load the trajectory
    trajectory = mdt.load(input_trajectory_filename, top=input_structure_filename)
    # Run a RMSD analysis
    rmsd_data = mdt.rmsd(trajectory, trajectory, 0)
    # Find sudden jumps in RMSD values
    sudden_jumps = []
    previous_value = 0
    for i, value in enumerate(rmsd_data):
        diff = value - previous_value
        # In case this is a new trajectory the jump will be negative
        if diff < cutoff:
            print('New trajectory at frame ' + str(i))
            sudden_jumps.append(i)
        previous_value = value
    # In there was no jumps then stop here
    if len(sudden_jumps) == 0:
        print('Apparently there is a single trajectory')
        return
    # Generate a trajectory subset for each cut
    cut_points = [ 0, *sudden_jumps, len(rmsd_data) ]
    for i in range(len(cut_points) -1):
        start = cut_points[i]
        end = cut_points[i+1]
        trajectory_split = trajectory[start:end]
        split_filename = output_trajectory_prefix + '_' + str(i+1) + '.' + input_trajectory_format
        print('Writting from frame ' + str(start) + ' to frame ' + str(end) + ' to "' + split_filename + '"')
        trajectory_split.save(split_filename)
    



# Main handler of the toolbelt
import os
from bisect import bisect
from typing import Optional, Tuple, List, Union
Coords = Tuple[float, float, float]

import prody

from .selections import Selection

# ------------------------------------------------------------------------------------

# Functions for type handling

# Get the residue and the residue index out of both the residue or the residue index
def parse_residue (input_residue : Union['Residue', int]) -> Tuple['Residue', int]:
    if type(input_residue) == int:
        residue_index = input_residue
        residue = residue.structure.residues[residue_index]
    elif type(input_residue) == 'Residue':
        residue = input_residue
        residue_index = residue.index
    else:
        raise ValueError('Unknow type when expecting Residue or int')
    return residue, residue_index

# Get the chain and the chain index out of both the chain or the chain index
def parse_chain (input_chain : Union['Chain', int]) -> Tuple['Chain', int]:
    if type(input_chain) == int:
        chain_index = input_chain
        chain = chain.structure.chains[chain_index]
    elif type(input_chain) == 'Chain':
        chain = input_chain
        chain_index = chain.index
    else:
        raise ValueError('Unknow type when expecting Chain or int')
    return chain, chain_index

# ------------------------------------------------------------------------------------

# An atom
class Atom:
    def __init__ (self,
        name : Optional[str] = None,
        element : Optional[str] = None,
        coords : Optional[Coords] = None,
        ):
        self.name = name
        self.element = element
        self.coords = coords
        # Set variables to store references to other related instances
        # These variables will be set further by the structure
        self.structure = None
        self.index = None
        self.residue = None
        self.chain = None
        self.residue_index = None
        self.chain_index = None

    def __repr__ (self):
        return '<Atom ' + self.name + '>'


# A residue
class Residue:
    def __init__ (self,
        name : Optional[str] = None,
        number : Optional[int] = None,
        icode : Optional[str] = None,
        atom_indices : List[int] = [],
        ):
        self.name = name
        self.number = number
        self.icode = icode
        self.atom_indices = atom_indices
        # Set variables to store references to other related instaces
        # These variables will be set further by the structure
        self.structure = None
        self.index = None
        self.atoms = []
        self.chain = None
        self.chain_index = None

    def __repr__ (self):
        return '<Residue ' + self.name + str(self.number) + (self.icode if self.icode else '') + '>'

    # Change the chain of the residue
    def change_chain (self, new_chain : Union['Chain', int]):
        # Get the residue current chain and remove the current residue from it
        current_chain = self.chain
        current_chain.remove_residue(self)
        # Get the new chain to be set
        new_chain, new_chain_index = parse_chain(new_chain)
        new_chain.add_residue(self)

# A chain
class Chain:
    def __init__ (self,
        name : Optional[str] = None,
        residue_indices : List[int] = [],
        ):
        self.name = name
        self.residue_indices = residue_indices
        # Set variables to store references to other related instaces
        # These variables will be set further by the structure
        self.structure = None
        self.index = None
        self.residues = []
        self.atoms = []
        self.atom_indices = []

    def __repr__ (self):
        return '<Chain ' + self.name + '>'

    # Add a residue to the chain
    def add_residue (self, input_residue : Union['Residue', int]):
        residue, residue_index = parse_residue(input_residue)
        sorted_residue_index = bisect(self.residue_indices, residue_index)
        self.residue_indices.insert(sorted_residue_index, residue_index)
        self.residues.insert(sorted_residue_index, residue)
        for atom_index in residue.atom_indices:
            sorted_atom_index = bisect(self.atom_indices, atom_index)
            self.atom_indices.insert(sorted_atom_index, atom_index)
            atom = self.structure.atoms[atom_index]
            self.atoms.insert(sorted_atom_index, atom)
            # Update atoms themselves
            atom.chain = self
            atom.chain_index = self.index
        # Update the residue itself
        residue.chain = self
        residue.chain_index = self.index

    # Remove a residue from the chain
    def remove_residue (self, input_residue : Union['Residue', int]):
        residue, residue_index = parse_residue(input_residue)
        self.residue_indices.remove(residue_index) # This index MUST be in the list
        self.residues.pop(residue_index)
        for atom_index in residue.atom_indices:
            self.atom_indices.remove(atom_index) # This index MUST be in the list
            self.atoms.pop(atom_index)
            # Update atoms themselves
            atom = self.structure.atoms[atom_index]
            atom.chain = None
            atom.chain_index = None
        # Update the residue itself
        residue.chain = None
        residue.chain_index = None

# A structure is a group of atoms organized in chains and residues
class Structure:
    def __init__ (self,
        atoms : List['Atom'] = [],
        residues : List['Residue'] = [],
        chains : List['Chain'] = [],
        ):
        self.atoms = []
        self.residues = []
        self.chains = []
        # Set references between instances
        for atom in atoms:
            self.set_new_atom(atom)
        for residue in residues:
            self.set_new_residue(residue)
        for chain in chains:
            self.set_new_chain(chain)

    def __repr__ (self):
        return '<Structure (' + str(len(self.atoms)) + ' atoms)>'

    # Set a new atom in the structure
    def set_new_atom (self, atom : 'Atom'):
        atom.structure = self
        new_atom_index = len(self.atoms)
        self.atoms.append(atom)
        atom.index = new_atom_index

    # Set a new residue in the structure
    # WARNING: Atoms must be set already before setting residues
    def set_new_residue (self, residue : 'Residue'):
        residue.structure = self
        new_residue_index = len(self.residues)
        self.residues.append(residue)
        residue.index = new_residue_index
        for atom_index in residue.atom_indices:
            atom = self.atoms[atom_index]
            residue.atoms.append(atom)
            atom.residue_index = new_residue_index
            atom.residue = residue

    # Set a new chain in the structure
    # WARNING: Residues and atoms must be set already before setting chains
    def set_new_chain (self, chain : 'Chain'):
        chain.structure = self
        new_chain_index = len(self.chains)
        self.chains.append(chain)
        chain.index = new_chain_index
        # Set chain residues and update those residues about the chain they belong to
        for residue_index in chain.residue_indices:
            residue = self.residues[residue_index]
            chain.residues.append(residue)
            residue.chain_index = new_chain_index
            residue.chain = chain
            # Set chain atoms and update those atoms about the chain they belong to
            for atom_index in residue.atom_indices:
                atom = self.atoms[atom_index]
                chain.atoms.append(atom)
                atom.chain_index = new_chain_index
                atom.chain = chain

    # Set the structure from a ProDy topology
    @classmethod
    def from_prody(cls, prody_topology):
        parsed_atoms = []
        parsed_residues = []
        parsed_chains = []
        prody_atoms = list(prody_topology.iterAtoms())
        prody_residues = list(prody_topology.iterResidues())
        prody_chains = list(prody_topology.iterChains())
        # Parse atoms
        for prody_atom in prody_atoms:
            name = prody_atom.getName()
            element = prody_atom.getElement()
            coords = tuple(prody_atom.getCoords())
            parsed_atom = Atom(name=name, element=element, coords=coords)
            parsed_atoms.append(parsed_atom)
        # Parse residues
        for prody_residue in prody_residues:
            name = prody_residue.getResname()
            number = int(prody_residue.getResnum())
            icode = prody_residue.getIcode()
            atom_indices = [ int(index) for index in prody_residue.getIndices() ]
            parsed_residue = Residue(name=name, number=number, icode=icode, atom_indices=atom_indices)
            parsed_residues.append(parsed_residue)
        # Parse chains
        for prody_chain in prody_chains:
            name = prody_chain.getChid()
            residue_indices = [ int(residue.getResindex()) for residue in prody_chain.iterResidues() ]
            parsed_chain = Chain(name=name, residue_indices=residue_indices)
            parsed_chains.append(parsed_chain)
        return cls(atoms=parsed_atoms, residues=parsed_residues, chains=parsed_chains)

    # Set the structure from a pdb file
    # Use ProDy to do so
    @classmethod
    def from_pdb_file(cls, pdb_filename : str):
        if not os.path.exists(pdb_filename):
            raise SystemExit('File "' + pdb_filename + '" not found')
        prody_topology = prody.parsePDB(pdb_filename)
        return cls.from_prody(prody_topology)

    # Generate a pdb file with current structure
    def generate_pdb_file(self, pdb_filename : str):
        with open(pdb_filename, "w") as file:
            file.write('REMARK mdtoolbelt dummy pdb file\n')
            for a, atom in enumerate(self.atoms):
                residue = atom.residue
                index = str(a+1).rjust(5)
                name =  ' ' + atom.name.ljust(3) if len(atom.name) < 4 else atom.name
                residue_name = residue.name.ljust(4)
                chain = atom.chain.name.rjust(1)
                residue_number = str(residue.number).rjust(4)
                icode = residue.icode.rjust(1)
                coords = atom.coords
                x_coord, y_coord, z_coord = [ "{:.3f}".format(coord).rjust(8) for coord in coords ]
                occupancy = '1.00' # Just a placeholder
                temp_factor = '0.00' # Just a placeholder
                element = atom.element
                atom_line = ('ATOM  ' + index + ' ' + name + ' ' + residue_name
                    + chain + residue_number + icode + '   ' + x_coord + y_coord + z_coord
                    + '  ' + occupancy + '  ' + temp_factor + ' ' + element).ljust(80) + '\n'
                file.write(atom_line)

    # Get the structure equivalent prody topology
    def get_prody_topology (self):
        # Generate the prody topology
        pdb_filename = '.structure.pdb'
        self.generate_pdb_file(pdb_filename)
        prody_topology = prody.parsePDB(pdb_filename)
        os.remove(pdb_filename)
        return prody_topology

    # Select atoms from the structure thus generating an atom indices list
    # Different tools may be used to make the selection:
    # - prody (default)
    def select (self, selection_string : str, logic : str = 'prody') -> Optional['Selection']:
        if logic == 'prody':
            prody_topology = self.get_prody_topology()
            prody_selection = prody_topology.select(selection_string)
            if not prody_selection:
                print('WARNING: Empty selection')
                return None
            return Selection.from_prody(prody_selection)
    
    # Create a new structure from the current using a selection to filter atoms
    def filter (self, selection : 'Selection') -> 'Structure':
        if not selection:
            raise SystemExit('No selection was passed')
        new_atoms = []
        new_residues = []
        new_chains = []
        # Get the selected atoms
        for index in selection.atom_indices:
            # Make a copy of the selected atoms in order to not modify the original ones
            original_atom = self.atoms[index]
            new_atom = Atom(
                name=original_atom.name,
                element=original_atom.element,
                coords=original_atom.coords,
                residue_index=original_atom.residue_index,
                chain_index=original_atom.chain_index
            )
            new_atoms.append(new_atom)
        # Find the selected residues
        selected_residue_indices = list(set([ atom.residue_index for atom in new_atoms ]))
        for index in selected_residue_indices:
            original_residue = self.residues[index]
            new_residue = Residue(
                name=original_residue.name,
                number=original_residue.number,
                icode=original_residue.icode,
                atom_indices=original_residue.atom_indices,
                chain_index=original_residue.chain_index
            )
            new_residues.append(new_residue)
        # Find the selected chains
        selected_chain_indices = list(set([ atom.chain_index for atom in new_atoms ]))
        for index in selected_chain_indices:
            original_chain = self.chains[index]
            new_chain = Chain(
                name=original_chain.name,
                atom_indices=original_chain.atom_indices,
                residue_indices=original_chain.residue_indices
            )
            new_chains.append(new_chain)
        # Generate dictionaries with new indexes as keys and previous indexes as values for atoms, residues and chains
        old_atom_indices = {}
        for i, index in enumerate(selection.atom_indices):
            old_atom_indices[index] = i
        old_residue_indices = {}
        for i, index in enumerate(selected_residue_indices):
            old_residue_indices[index] = i
        old_chain_indices = {}
        for i, index in enumerate(selected_chain_indices):
            old_chain_indices[index] = i
        # Finally, reset indices in all instances
        for atom in new_atoms:
            atom.residue_index = old_residue_indices[atom.residue_index]
            atom.chain_index = old_chain_indices[atom.chain_index]
        for residue in new_residues:
            residue.atom_indices = [ old_atom_indices[index] for index in residue.atom_indices ]
            residue.chain_index = old_chain_indices[residue.chain_index]
        for chain in new_chains:
            chain.atom_indices = [ old_atom_indices[index] for index in chain.atom_indices ]
            chain.residue_indices = [ old_residue_indices[index] for index in chain.residue_indices ]
        return Structure(atoms=new_atoms, residues=new_residues, chains=new_chains)

    # Set chains on demand
    # If no selection is passed then the whole structure will be affected
    # If no chain is passed then a "chain by fragment" logic will be applied
    def chainer (self, selection : Optional['Selection'] = None, letter : Optional[str] = None):

        # DANI: Aquí falta encontrar los strong bonds, para lo cual hay que extraer la lógica del workflow

        # DANI: Si la cadena no existe habrá que crearla
        #     new_chain = Chain(name=new_chain.name, residue_indices=[self.index])
        #     self.structure.set_new_chain(new_chain) # This function updates atoms and this residue already

        # DANI: Cuando el cambio esté claro para cada residuo:
        #     residue.change_chain(new_chain)

        pass

    # Get a chain by its name
    def get_chain_by_name (self, name : str) -> 'Chain':
        return next((c for c in self.chains if c.name == name), None)


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='mdtoolbelt',
    version='0.0.3',
    author="Dani Beltrán",
    author_email="d.beltran.anadon@gmail.com",
    description="Tools por MD post-processing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/d-beltran/mdtoolbelt",
    project_urls={
        "Bug Tracker": "https://github.com/d-beltran/mdtoolbelt/issues",
    },
    packages=['mdtoolbelt'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={
        'console_scripts': [
            'mdtb = mdtoolbelt.console:call'
        ]
    },
    python_requires=">=3.6",
)